from wpimath.geometry import Rotation2d, Translation2d
from wpiutil import SendableBuilder

from ..abstract.motor import CoaxialDriveComponent, CoaxialAzimuthComponent
from ..abstract.system import SwerveModule


class CoaxialSwerveModule(SwerveModule):
    last_commanded_drive_velocity: float = 0
    last_commanded_azimuth_angle = Rotation2d.fromDegrees(0)

    def __init__(self, drive: CoaxialDriveComponent, azimuth: CoaxialAzimuthComponent, placement: Translation2d):
        super().__init__()

        self._drive = drive
        self._azimuth = azimuth
        self.placement = placement

    def desire_drive_velocity(self, velocity: float, open_loop: bool):
        self.last_commanded_drive_velocity = velocity
        if open_loop:
            self._drive.follow_velocity_open(velocity)
        else:
            self._drive.follow_velocity_closed(velocity)

    def desire_azimuth_angle(self, angle: Rotation2d):
        self.last_commanded_azimuth_angle = angle
        self._azimuth.follow_angle(angle)

    def reset(self):
        self._drive.reset()
        self._azimuth.reset()

    @property
    def drive_velocity(self) -> float:
        return self._drive.velocity

    @property
    def drive_distance(self) -> float:
        return self._drive.distance

    @property
    def azimuth_angle(self) -> Rotation2d:
        return self._azimuth.angle

    @property
    def azimuth_velocity(self) -> float:
        return self._azimuth.rotational_velocity

    def initSendable(self, builder: SendableBuilder):
        # fmt: off
        builder.setSmartDashboardType("CoaxialSwerveModule")
        builder.addDoubleProperty("Drive Velocity (mps)", lambda: self._drive.velocity, lambda _: None)
        builder.addDoubleProperty("Drive Distance (m)", lambda: self._drive.distance, lambda _: None)
        builder.addDoubleProperty("Azimuth Velocity (radps)", lambda: self._azimuth.rotational_velocity, lambda _: None)
        builder.addDoubleProperty("Azimuth Position (rad)", lambda: self._azimuth.angle.radians(), lambda _: None)
        builder.addDoubleProperty("Azimuth Position (deg)", lambda: self._azimuth.angle.degrees(), lambda _: None)
        builder.addDoubleProperty("Desired Drive Velocity (mps)", lambda: self.last_commanded_drive_velocity, lambda _: None)
        builder.addDoubleProperty("Desired Azimuth Position (rad)", lambda: self.last_commanded_azimuth_angle.radians(), lambda _: None)
        builder.addDoubleProperty("Desired Azimuth Position (deg)", lambda: self.last_commanded_azimuth_angle.degrees(), lambda _: None)
        # fmt: on



from abc import abstractmethod

from wpimath.geometry import Translation2d, Rotation2d
from wpimath.kinematics import SwerveModuleState, SwerveModulePosition
from wpiutil import Sendable

from . import SendableABCMeta


class SwerveModule(Sendable, metaclass=SendableABCMeta):
    placement: Translation2d

    def desire_state(self, state: SwerveModuleState, drive_open_loop: bool, rotate_in_place: bool):
        """
        Command the module to a speed and angle

        :param state: SwerveModuleState representing the module's desired speed and angle
        :param drive_open_loop: Use open loop control (True) or closed loop (False) to drive the wheel
        :param rotate_in_place: Should the modules rotate while not driving
        :return:
        """
        state = optimize(state, self.azimuth_angle)

        # Prevent rotating the module if drive speed is less than 2 cm/s to prevent feedback-loop jitter
        angle = state.angle if rotate_in_place or abs(state.speed) > 0.02 else self.azimuth_angle

        self.desire_drive_velocity(state.speed, drive_open_loop)
        self.desire_azimuth_angle(angle)

    @property
    def module_position(self) -> SwerveModulePosition:
        return SwerveModulePosition(self.drive_distance, self.azimuth_angle)

    @abstractmethod
    def desire_drive_velocity(self, velocity: float, open_loop: bool):
        """
        Drive the wheel

        :param velocity: Desired velocity in m/s
        :param open_loop: Use open loop control (True) or closed loop (False)
        """
        raise NotImplementedError

    @abstractmethod
    def desire_azimuth_angle(self, angle: Rotation2d):
        raise NotImplementedError

    @abstractmethod
    def reset(self):
        raise NotImplementedError

    @property
    @abstractmethod
    def drive_velocity(self) -> float:
        """Drive wheel velocity in m/s"""
        raise NotImplementedError

    @property
    @abstractmethod
    def drive_distance(self) -> float:
        """Driven distance in metres"""
        raise NotImplementedError

    @property
    @abstractmethod
    def azimuth_angle(self) -> Rotation2d:
        """CCW+ module angle"""
        raise NotImplementedError

    @property
    @abstractmethod
    def azimuth_velocity(self) -> float:
        """CCW+ module angular velocity in rad/s"""
        raise NotImplementedError


def sign(num):
    return 1 if num > 0 else -1 if num < 0 else 0


def place_in_proper_0_to_360_scope(scope_reference: float, new_angle: float) -> float:
    # Place the new_angle in the range that is a multiple of [0, 360] (e.g., [360, 720]) which is closest
    # to the scope_reference
    lower_offset = scope_reference % 360
    lower_bound = scope_reference - lower_offset
    upper_bound = lower_bound + 360

    while new_angle < lower_bound:
        new_angle += 360
    while new_angle > upper_bound:
        new_angle -= 360

    if new_angle - scope_reference > 180:
        new_angle -= 360
    elif new_angle - scope_reference < -180:
        new_angle += 360

    return new_angle


def optimize(desired_state: SwerveModuleState, current_angle: Rotation2d):
    # There are two ways for a swerve module to reach its goal
    # 1) Rotate to its intended rotation and drive at its intended speed
    # 2) Rotate to the mirrored rotation (subtract 180) and drive at the opposite of its intended speed
    # Optimizing finds the option that requires the smallest rotation by the module

    target_angle = place_in_proper_0_to_360_scope(current_angle.degrees(), desired_state.angle.degrees())
    target_speed = desired_state.speed
    delta = target_angle - current_angle.degrees()

    if abs(delta) > 90:
        target_speed *= -1
        target_angle -= 180 * sign(delta)

    return SwerveModuleState(target_speed, Rotation2d.fromDegrees(target_angle))



"""
A modular swerve drive library extendable to any swerve configuration.
Comes with support for coaxial SDS- and MAXSwerve-style drivetrains out of the box.
"""

__all__ = ["u", "SwerveDrive", "TrajectoryFollowerParameters"]

# fmt: off

# Initialize the unit registry before importing anything that relies on it
from pint import UnitRegistry
u = UnitRegistry()

from .subsystem import SwerveDrive, TrajectoryFollowerParameters



from abc import abstractmethod

from wpimath.geometry import Rotation2d
from wpiutil import Sendable, SendableBuilder

from . import SendableABCMeta


class Gyro(Sendable, metaclass=SendableABCMeta):
    @abstractmethod
    def zero_heading(self):
        raise NotImplementedError

    @property
    @abstractmethod
    def heading(self) -> Rotation2d:
        """CCW+ chassis yaw angle"""
        raise NotImplementedError

    def initSendable(self, builder: SendableBuilder):
        builder.setSmartDashboardType("Gyro")
        builder.addDoubleProperty("Value", lambda: self.heading.degrees(), lambda _: None)
        builder.addDoubleProperty("Heading (rad)", lambda: self.heading.radians(), lambda _: None)


class AbsoluteEncoder(Sendable, metaclass=SendableABCMeta):
    @property
    @abstractmethod
    def absolute_position(self) -> Rotation2d:
        raise NotImplementedError

    def initSendable(self, builder: SendableBuilder):
        builder.setSmartDashboardType("AbsoluteEncoder")
        builder.addDoubleProperty("Absolute Rotation (rad)", lambda: self.absolute_position.radians(), lambda _: None)
        builder.addDoubleProperty("Absolute Rotation (deg)", lambda: self.absolute_position.degrees(), lambda _: None)



"""The swerve drive subsystem and other classes it relies on"""

import math
import time
from dataclasses import dataclass
from functools import cached_property
from typing import Callable, Optional, TYPE_CHECKING

import commands2
import wpilib
import wpimath.estimator
import wpimath.kinematics
from pint import Quantity
from wpimath.controller import ProfiledPIDControllerRadians, PIDController
from wpimath.trajectory import Trajectory, TrapezoidProfileRadians
from wpimath.geometry import Pose2d, Translation2d
from wpimath.kinematics import ChassisSpeeds, SwerveModuleState, SwerveModulePosition
from wpiutil import SendableBuilder

if TYPE_CHECKING:
    from wpimath.estimator import SwerveDrive4PoseEstimator
    from wpimath.kinematics import SwerveDrive4Kinematics

from swervepy import u
from swervepy.abstract import SwerveModule, Gyro


class SwerveDrive(commands2.SubsystemBase):
    def __init__(
        self,
        modules: tuple[SwerveModule, ...],
        gyro: Gyro,
        max_velocity: Quantity,
        max_angular_velocity: Quantity,
        vision_pose_callback: Callable[[Pose2d], Optional[Pose2d]] = lambda _: None,
    ):
        super().__init__()

        self._modules = modules
        self._gyro = gyro
        self._vision_pose_callback = vision_pose_callback
        self.max_velocity: float = max_velocity.m_as(u.m / u.s)
        self.max_angular_velocity: float = max_angular_velocity.m_as(u.rad / u.s)

        # Pause init for a second before setting module offsets to avoid a bug related to inverting motors.
        # Fixes https://github.com/Team364/BaseFalconSwerve/issues/8.
        time.sleep(1)
        self.reset_modules()

        # Zero heading at startup to set "forward" direction
        self.zero_heading()

        # There are different classes for each number of swerve modules in a drive base,
        # so construct the class name from number of modules.
        self._kinematics: "SwerveDrive4Kinematics" = getattr(
            wpimath.kinematics, f"SwerveDrive{len(modules)}Kinematics"
        )(*[module.placement for module in self._modules])
        self._odometry: "SwerveDrive4PoseEstimator" = getattr(
            wpimath.estimator, f"SwerveDrive{len(modules)}PoseEstimator"
        )(self._kinematics, self._gyro.heading, self.module_positions, Pose2d())

        for i, module in enumerate(modules):
            wpilib.SmartDashboard.putData(f"Module {i}", module)

        # Field to plot auto trajectories and robot pose
        self.field = wpilib.Field2d()
        wpilib.SmartDashboard.putData(self.field)

    def periodic(self):
        robot_pose = self._odometry.update(self._gyro.heading, self.module_positions)

        vision_pose = self._vision_pose_callback(self.pose)
        if vision_pose:
            self._odometry.addVisionMeasurement(vision_pose, wpilib.Timer.getFPGATimestamp())

        # Visualize robot position on field
        self.field.setRobotPose(robot_pose)

    def drive(self, translation: Translation2d, rotation: float, field_relative: bool, open_loop: bool):
        """
        Command the robot to provided chassis speeds (translation and rotation)

        :param translation: Translation speed on the XY-plane in m/s where +X is forward and +Y is left
        :param rotation: Rotation speed around the Z-axis in rad/s where CCW+
        :param field_relative: If True, gyroscopic zero is used as the forward direction.
        Else, forward faces the front of the robot.
        :param open_loop: Use open loop control (True) or closed loop (False)
        """

        speeds = (
            ChassisSpeeds.fromFieldRelativeSpeeds(translation.x, translation.y, rotation, self._gyro.heading)
            if field_relative
            else ChassisSpeeds(translation.x, translation.y, rotation)
        )
        swerve_module_states = self._kinematics.toSwerveModuleStates(speeds)

        self.desire_module_states(swerve_module_states, open_loop, rotate_in_place=False)

    def desire_module_states(
        self, states: tuple[SwerveModuleState, ...], open_loop: bool = False, rotate_in_place: bool = True
    ):
        """
        Command each individual module to a state

        :param states: List of module states in the order of the swerve module list SwerveDrive was created with
        :param open_loop: Use open loop control (True) or closed loop (False)
        :param rotate_in_place: Should the modules rotate while not driving
        """

        swerve_module_states = self._kinematics.desaturateWheelSpeeds(states, self.max_velocity)  # type: ignore

        for i in range(4):
            module: SwerveModule = self._modules[i]
            module.desire_state(swerve_module_states[i], open_loop, rotate_in_place)

    @property
    def module_positions(self) -> tuple[SwerveModulePosition, ...]:
        return tuple(module.module_position for module in self._modules)

    @property
    def pose(self) -> Pose2d:
        return self._odometry.getEstimatedPosition()

    def reset_modules(self):
        for module in self._modules:
            module.reset()

    def zero_heading(self):
        self._gyro.zero_heading()

    def reset_odometry(self, pose: Pose2d):
        """
        Reset the drive base's pose to a new one

        :param pose: The new pose
        """

        self._odometry.resetPosition(self._gyro.heading, self.module_positions, pose)  # type: ignore

    def teleop_command(
        self,
        translation: Callable[[], float],
        strafe: Callable[[], float],
        rotation: Callable[[], float],
        field_relative: bool,
        open_loop: bool,
    ):
        return _TeleOpCommand(self, translation, strafe, rotation, field_relative, open_loop)

    def follow_trajectory_command(
        self, trajectory: Trajectory, parameters: "TrajectoryFollowerParameters", first_path: bool = False
    ):
        """
        Build a command that follows a trajectory

        :param trajectory: The path to follow
        :param parameters: Options that determine how the robot will follow the trajectory
        :param first_path: If True, the robot's pose will be reset to the trajectory's initial pose
        :return: Trajectory-follower command
        """

        theta_controller = ProfiledPIDControllerRadians(
            parameters.theta_kP, 0, 0, parameters.theta_controller_constraints
        )
        theta_controller.enableContinuousInput(-math.pi, math.pi)

        command = commands2.Swerve4ControllerCommand(
            trajectory,
            lambda: self.pose,
            self._kinematics,
            PIDController(parameters.x_kP, 0, 0),
            PIDController(parameters.y_kP, 0, 0),
            theta_controller,
            self.desire_module_states,
            [self],
        ).beforeStarting(lambda: self.field.getObject("traj").setTrajectory(trajectory))

        # If this is the first path in a sequence, reset the robot's pose so that it aligns with the start of the path
        if first_path:
            initial_pose = trajectory.initialPose()
            command = command.beforeStarting(lambda: self.reset_odometry(initial_pose))

        return command


class _TeleOpCommand(commands2.CommandBase):
    def __init__(
        self,
        swerve: SwerveDrive,
        translation: Callable[[], float],
        strafe: Callable[[], float],
        rotation: Callable[[], float],
        field_relative: bool,
        open_loop: bool,
    ):
        super().__init__()
        self.addRequirements(swerve)

        self._swerve = swerve
        self._translation = translation
        self._strafe = strafe
        self._rotation = rotation
        self.field_relative = field_relative
        self.open_loop = open_loop

    def execute(self):
        self._swerve.drive(
            Translation2d(self._translation(), self._strafe()) * self._swerve.max_velocity,
            self._rotation() * self._swerve.max_angular_velocity,
            self.field_relative,
            self.open_loop,
        )

    def initSendable(self, builder: SendableBuilder):
        # fmt: off
        builder.addBooleanProperty("Field Relative", lambda: self.field_relative,
                                   lambda val: setattr(self, "field_relative", val))
        builder.addBooleanProperty("Open Loop", lambda: self.open_loop, lambda val: setattr(self, "open_loop", val))
        # fmt: on

    def toggle_field_relative(self):
        self.field_relative = not self.field_relative

    def toggle_open_loop(self):
        self.open_loop = not self.open_loop


@dataclass
class TrajectoryFollowerParameters:
    target_angular_velocity: Quantity
    target_angular_acceleration: Quantity

    # Positional PID constants for X, Y, and theta (rotation) controllers
    theta_kP: float
    x_kP: float
    y_kP: float

    @cached_property
    def theta_controller_constraints(self):
        return TrapezoidProfileRadians.Constraints(
            self.target_angular_velocity.m_as(u.rad / u.s),
            self.target_angular_acceleration.m_as(u.rad / (u.s * u.s)),
        )



"""
A collection of methods for converting between native Falcon 500 units and standard units, like metres.
"""
import math

import wpimath.geometry

FALCON_CPR = 2048
DEGREES_PER_ROTATION = 360
RADS_PER_ROTATION = 2 * math.pi


def falcon_to_degrees(counts: float, gear_ratio: float) -> wpimath.geometry.Rotation2d:
    degrees = counts * (DEGREES_PER_ROTATION / (gear_ratio * FALCON_CPR))
    return wpimath.geometry.Rotation2d.fromDegrees(degrees)


def degrees_to_falcon(degrees: wpimath.geometry.Rotation2d, gear_ratio: float) -> float:
    ticks = degrees.degrees() / (DEGREES_PER_ROTATION / (gear_ratio * FALCON_CPR))
    return ticks


def falcon_to_rpm(velocity: float, gear_ratio: float) -> float:
    # Falcon velocity is in units/100ms, so multiplying by 600 changes to units/min.
    motor_rpm = velocity * (600 / FALCON_CPR)
    mechanism_rpm = motor_rpm / gear_ratio
    return mechanism_rpm


def rpm_to_falcon(rpm: float, gear_ratio: float) -> float:
    motor_rpm = rpm * gear_ratio
    # Falcon velocity is in units/100ms, so multiplying by 600 changes to units/min.
    ticks = motor_rpm * (FALCON_CPR / 600)
    return ticks


def falcon_to_mps(velocity: float, circumference: float, gear_ratio: float) -> float:
    wheel_rpm = falcon_to_rpm(velocity, gear_ratio)
    # Divide by 60 to change from m/min to m/s
    wheel_mps = (wheel_rpm * circumference) / 60
    return wheel_mps


def mps_to_falcon(velocity: float, circumference: float, gear_ratio: float) -> float:
    # Multiply by 60 to change m/s to m/min
    wheel_rpm = (velocity * 60) / circumference
    wheel_velocity = rpm_to_falcon(wheel_rpm, gear_ratio)
    return wheel_velocity


def falcon_to_metres(counts: float, circumference: float, gear_ratio: float) -> float:
    rotations = counts / (gear_ratio * FALCON_CPR)
    metres = rotations * circumference
    return metres


def metres_to_falcon(metres: float, circumference: float, gear_ratio: float) -> float:
    rotations = metres / circumference
    counts = rotations * (gear_ratio * FALCON_CPR)
    return counts


def falcon_to_dps(velocity: float, gear_ratio: float) -> float:
    return falcon_to_rpm(velocity, gear_ratio) * DEGREES_PER_ROTATION / 60


def dps_to_falcon(dps: float, gear_ratio: float) -> float:
    rpm = dps * 60 / DEGREES_PER_ROTATION
    return rpm_to_falcon(rpm, gear_ratio)


def falcon_to_radps(velocity: float, gear_ratio: float) -> float:
    return falcon_to_rpm(velocity, gear_ratio) * RADS_PER_ROTATION / 60


def radps_to_falcon(dps: float, gear_ratio: float) -> float:
    rpm = dps * 60 / RADS_PER_ROTATION
    return rpm_to_falcon(rpm, gear_ratio)



from dataclasses import dataclass

import photonvision
import wpimath.geometry
import robotpy_apriltag as apriltag


@dataclass
class CameraDefinition:
    name: str
    position: wpimath.geometry.Transform3d


class AprilTagCameraCollection:
    def __init__(self, camera_definitions: list[CameraDefinition], field_layout: apriltag.AprilTagFieldLayout):
        cameras = [
            (photonvision.PhotonCamera(definition.name), definition.position) for definition in camera_definitions
        ]
        self.pose_estimator = photonvision.RobotPoseEstimator(
            field_layout, photonvision.PoseStrategy.CLOSEST_TO_REFERENCE_POSE, cameras
        )

    def estimate_pose(self, previous_estimated_pose: wpimath.geometry.Pose2d) -> wpimath.geometry.Pose2d:
        self.pose_estimator.setReferencePose(wpimath.geometry.Pose3d(previous_estimated_pose))
        # The update() method returns a tuple with an estimated pose as its 0th element and a timestamp as its 1st.
        new_pose = self.pose_estimator.update()[0]
        return new_pose.toPose2d()



"""
Contains default implementations of components (motors, sensors, modules). The user should instantiate these when
creating their drive base.
"""

__all__ = [
    "Falcon500CoaxialAzimuthComponent",
    "Falcon500CoaxialDriveComponent",
    "NEOCoaxialAzimuthComponent",
    "NEOCoaxialDriveComponent",
    "NEOCoaxialAzimuthComponent",
    "AbsoluteCANCoder",
    "PigeonGyro",
    "CoaxialSwerveModule",
    "SparkMaxEncoderType",
    "SparkMaxAbsoluteEncoder",
    "DummyCoaxialComponent",
    "DummyGyro",
]

from .motor import (
    Falcon500CoaxialAzimuthComponent,
    Falcon500CoaxialDriveComponent,
    NEOCoaxialDriveComponent,
    NEOCoaxialAzimuthComponent,
    NEOCoaxialAzimuthComponent,
    DummyCoaxialComponent,
)
from .sensor import AbsoluteCANCoder, PigeonGyro, SparkMaxEncoderType, SparkMaxAbsoluteEncoder, DummyGyro
from .system import CoaxialSwerveModule



from abc import abstractmethod
from typing import Protocol

from wpimath.geometry import Rotation2d


class CoaxialDriveComponent(Protocol):
    """Drive motor and encoder"""

    @abstractmethod
    def follow_velocity_open(self, velocity: float):
        """
        Follow a velocity using open loop control

        :param velocity: Desired velocity in m/s
        """
        raise NotImplementedError

    @abstractmethod
    def follow_velocity_closed(self, velocity: float):
        """
        Follow a velocity using closed loop control (i.e., PID)

        :param velocity: Desired velocity in m/s
        """
        raise NotImplementedError

    @abstractmethod
    def reset(self):
        """Reset driven distance to zero or otherwise reinitialize the motor"""
        raise NotImplementedError

    @property
    @abstractmethod
    def velocity(self) -> float:
        """Drive velocity in m/s"""
        raise NotImplementedError

    @property
    @abstractmethod
    def distance(self) -> float:
        """Driven distance in metres"""
        raise NotImplementedError


class CoaxialAzimuthComponent(Protocol):
    """
    Azimuth (turning) motor and encoder.
    Also includes a method for resetting to an absolute position (often provided by an absolute encoder)
    """

    @abstractmethod
    def follow_angle(self, angle: Rotation2d):
        raise NotImplementedError

    @abstractmethod
    def reset(self):
        """Reset module angle to absolute position"""
        raise NotImplementedError

    @property
    @abstractmethod
    def rotational_velocity(self) -> float:
        """Rotational velocity in rad/s"""
        raise NotImplementedError

    @property
    @abstractmethod
    def angle(self) -> Rotation2d:
        raise NotImplementedError



"""
Contains interfaces for components used in a swerve drive base. These are sensors, motors, and the swerve module itself.
Implementations can be found in the impl module, or the user may define their own.
"""

__all__ = [
    "SendableABCMeta",
    "CoaxialDriveComponent",
    "CoaxialAzimuthComponent",
    "Gyro",
    "AbsoluteEncoder",
    "SwerveModule",
]

from abc import ABCMeta
from wpiutil import Sendable


class SendableABCMeta(ABCMeta, type(Sendable)):
    pass


from .motor import CoaxialDriveComponent, CoaxialAzimuthComponent
from .sensor import Gyro, AbsoluteEncoder
from .system import SwerveModule



import copy
from dataclasses import dataclass

import ctre
import rev
from pint import Quantity
from wpimath.controller import SimpleMotorFeedforwardMeters
from wpimath.geometry import Rotation2d

from .sensor import SparkMaxEncoderType, SparkMaxAbsoluteEncoder
from ..abstract.motor import CoaxialDriveComponent, CoaxialAzimuthComponent
from .. import conversions, u
from ..abstract.sensor import AbsoluteEncoder


class Falcon500CoaxialDriveComponent(CoaxialDriveComponent):
    @dataclass
    class Parameters:
        wheel_circumference: Quantity
        gear_ratio: float

        max_speed: Quantity

        open_loop_ramp_rate: float
        closed_loop_ramp_rate: float

        continuous_current_limit: int
        peak_current_limit: int
        peak_current_duration: float

        neutral_mode: ctre.NeutralMode

        kP: float
        kI: float
        kD: float

        kS: float
        kV: float
        kA: float

        invert_motor: bool

        def in_standard_units(self):
            data = copy.deepcopy(self)
            data.wheel_circumference = data.wheel_circumference.m_as(u.m)
            data.max_speed = data.max_speed.m_as(u.m / u.s)
            return data

        # noinspection PyPep8Naming
        def create_TalonFX_config(self) -> ctre.TalonFXConfiguration:
            motor_config = ctre.TalonFXConfiguration()

            supply_limit = ctre.SupplyCurrentLimitConfiguration(
                True,
                self.continuous_current_limit,
                self.peak_current_limit,
                self.peak_current_duration,
            )

            motor_config.slot0.kP = self.kP
            motor_config.slot0.kI = self.kI
            motor_config.slot0.kD = self.kD
            motor_config.supplyCurrLimit = supply_limit
            motor_config.initializationStrategy = ctre.sensors.SensorInitializationStrategy.BootToZero
            motor_config.openloopRamp = self.open_loop_ramp_rate
            motor_config.closedloopRamp = self.closed_loop_ramp_rate

            return motor_config

    def __init__(self, id_: int | tuple[int, str], parameters: Parameters):
        self._params = parameters.in_standard_units()

        try:
            # Unpack tuple of motor id and CAN bus id into TalonFX constructor
            self._motor = ctre.TalonFX(*id_)
        except TypeError:
            # Only an int was provided for id_
            self._motor = ctre.TalonFX(id_)

        self._config()
        self.reset()

        self._feedforward = SimpleMotorFeedforwardMeters(parameters.kS, parameters.kV, parameters.kA)

    def _config(self):
        settings = self._params.create_TalonFX_config()
        self._motor.configFactoryDefault()
        self._motor.configAllSettings(settings)
        self._motor.setInverted(self._params.invert_motor)
        self._motor.setNeutralMode(self._params.neutral_mode)

    def follow_velocity_open(self, velocity: float):
        percent_out = velocity / self._params.max_speed
        self._motor.set(ctre.ControlMode.PercentOutput, percent_out)

    def follow_velocity_closed(self, velocity: float):
        converted_velocity = conversions.mps_to_falcon(
            velocity, self._params.wheel_circumference, self._params.gear_ratio
        )
        self._motor.set(
            ctre.ControlMode.Velocity,
            converted_velocity,
            ctre.DemandType.ArbitraryFeedForward,
            self._feedforward.calculate(velocity),
        )

    def reset(self):
        self._motor.setSelectedSensorPosition(0)

    @property
    def velocity(self) -> float:
        return conversions.falcon_to_mps(
            self._motor.getSelectedSensorVelocity(),
            self._params.wheel_circumference,
            self._params.gear_ratio,
        )

    @property
    def distance(self) -> float:
        return conversions.falcon_to_metres(
            self._motor.getSelectedSensorPosition(),
            self._params.wheel_circumference,
            self._params.gear_ratio,
        )


class Falcon500CoaxialAzimuthComponent(CoaxialAzimuthComponent):
    @dataclass
    class Parameters:
        gear_ratio: float

        max_angular_velocity: Quantity

        ramp_rate: float

        continuous_current_limit: int
        peak_current_limit: int
        peak_current_duration: float

        neutral_mode: ctre.NeutralMode

        kP: float
        kI: float
        kD: float

        invert_motor: bool

        def in_standard_units(self):
            data = copy.deepcopy(self)
            data.max_angular_velocity = data.max_angular_velocity.m_as(u.rad / u.s)
            return data

        # noinspection PyPep8Naming
        def create_TalonFX_config(self) -> ctre.TalonFXConfiguration:
            motor_config = ctre.TalonFXConfiguration()

            supply_limit = ctre.SupplyCurrentLimitConfiguration(
                True,
                self.continuous_current_limit,
                self.peak_current_limit,
                self.peak_current_duration,
            )

            motor_config.slot0.kP = self.kP
            motor_config.slot0.kI = self.kI
            motor_config.slot0.kD = self.kD
            motor_config.supplyCurrLimit = supply_limit
            motor_config.initializationStrategy = ctre.sensors.SensorInitializationStrategy.BootToZero
            motor_config.closedloopRamp = self.ramp_rate

            return motor_config

    def __init__(
        self,
        id_: int | tuple[int, str],
        azimuth_offset: Rotation2d,
        parameters: Parameters,
        absolute_encoder: AbsoluteEncoder,
    ):
        self._params = parameters.in_standard_units()

        try:
            # Unpack tuple of motor id and CAN bus id into TalonFX constructor
            self._motor = ctre.TalonFX(*id_)
        except TypeError:
            # Only an int was provided for id_
            self._motor = ctre.TalonFX(id_)

        self._absolute_encoder = absolute_encoder
        self._offset = azimuth_offset

        self._config()
        self.reset()

    def _config(self):
        settings = self._params.create_TalonFX_config()
        self._motor.configFactoryDefault()
        self._motor.configAllSettings(settings)
        self._motor.setInverted(self._params.invert_motor)
        self._motor.setNeutralMode(self._params.neutral_mode)

    def follow_angle(self, angle: Rotation2d):
        converted_angle = conversions.degrees_to_falcon(angle, self._params.gear_ratio)
        self._motor.set(ctre.ControlMode.Position, converted_angle)

    def reset(self):
        absolute_position = self._absolute_encoder.absolute_position - self._offset
        converted_position = conversions.degrees_to_falcon(absolute_position, self._params.gear_ratio)
        self._motor.setSelectedSensorPosition(converted_position)

    @property
    def rotational_velocity(self) -> float:
        return conversions.falcon_to_radps(self._motor.getSelectedSensorVelocity(), self._params.gear_ratio)

    @property
    def angle(self) -> Rotation2d:
        return conversions.falcon_to_degrees(self._motor.getSelectedSensorPosition(), self._params.gear_ratio)


class NEOCoaxialDriveComponent(CoaxialDriveComponent):
    @dataclass
    class Parameters:
        wheel_circumference: Quantity
        gear_ratio: float

        max_speed: Quantity

        open_loop_ramp_rate: float
        closed_loop_ramp_rate: float

        continuous_current_limit: int
        peak_current_limit: int

        neutral_mode: rev.CANSparkMax.IdleMode

        kP: float
        kI: float
        kD: float

        kS: float
        kV: float
        kA: float

        invert_motor: bool

        def in_standard_units(self):
            data = copy.deepcopy(self)
            data.wheel_circumference = data.wheel_circumference.m_as(u.m)
            data.max_speed = data.max_speed.m_as(u.m / u.s)
            return data

    def __init__(self, id_: int, parameters: Parameters):
        self._params = parameters.in_standard_units()

        self._motor = rev.CANSparkMax(id_, rev.CANSparkMax.MotorType.kBrushless)
        self._controller = self._motor.getPIDController()
        self._encoder = self._motor.getEncoder()
        self._config()
        self.reset()

        self._feedforward = SimpleMotorFeedforwardMeters(parameters.kS, parameters.kV, parameters.kA)

    def _config(self):
        self._motor.restoreFactoryDefaults()

        self._controller.setP(self._params.kP)
        self._controller.setI(self._params.kI)
        self._controller.setD(self._params.kD)

        self._motor.setSmartCurrentLimit(self._params.continuous_current_limit)
        self._motor.setSecondaryCurrentLimit(self._params.peak_current_limit)

        self._motor.setOpenLoopRampRate(self._params.open_loop_ramp_rate)
        self._motor.setClosedLoopRampRate(self._params.closed_loop_ramp_rate)

        self._motor.setInverted(self._params.invert_motor)
        self._motor.setIdleMode(self._params.neutral_mode)

        position_conversion_factor = self._params.wheel_circumference / self._params.gear_ratio
        self._encoder.setPositionConversionFactor(position_conversion_factor)
        self._encoder.setVelocityConversionFactor(position_conversion_factor / 60)

    def follow_velocity_open(self, velocity: float):
        percent_out = velocity / self._params.max_speed
        self._motor.set(percent_out)

    def follow_velocity_closed(self, velocity: float):
        self._controller.setReference(
            velocity,
            rev.CANSparkMax.ControlType.kVelocity,
            arbFeedforward=self._feedforward.calculate(velocity),
        )

    def reset(self):
        self._encoder.setPosition(0)

    @property
    def velocity(self) -> float:
        return self._encoder.getVelocity()

    @property
    def distance(self) -> float:
        return self._encoder.getPosition()


class NEOCoaxialAzimuthComponent(CoaxialAzimuthComponent):
    @dataclass
    class Parameters:
        gear_ratio: float

        max_angular_velocity: Quantity

        ramp_rate: float

        continuous_current_limit: int
        peak_current_limit: int

        neutral_mode: rev.CANSparkMax.IdleMode

        kP: float
        kI: float
        kD: float

        invert_motor: bool

        def in_standard_units(self):
            data = copy.deepcopy(self)
            data.max_angular_velocity = data.max_angular_velocity.m_as(u.rad / u.s)
            return data

    def __init__(
        self,
        id_: int,
        azimuth_offset: Rotation2d,
        parameters: Parameters,
        absolute_encoder: AbsoluteEncoder | SparkMaxEncoderType,
    ):
        self._params = parameters.in_standard_units()

        self._motor = rev.CANSparkMax(id_, rev.CANSparkMax.MotorType.kBrushless)
        self._controller = self._motor.getPIDController()
        self._encoder = self._motor.getEncoder()

        # Config must be called before the absolute encoder is set up because config method
        # factory resets the SPARK MAX
        self._config()

        if isinstance(absolute_encoder, SparkMaxEncoderType):
            # Construct an AbsoluteEncoder from sensor plugged into SPARK MAX
            self._absolute_encoder = SparkMaxAbsoluteEncoder(self._motor, absolute_encoder)
        else:
            self._absolute_encoder = absolute_encoder

        self._offset = azimuth_offset

        self.reset()

    def _config(self):
        self._motor.restoreFactoryDefaults()

        self._controller.setP(self._params.kP)
        self._controller.setI(self._params.kI)
        self._controller.setD(self._params.kD)

        self._motor.setSmartCurrentLimit(self._params.continuous_current_limit)
        self._motor.setSecondaryCurrentLimit(self._params.peak_current_limit)

        self._motor.setInverted(self._params.invert_motor)
        self._motor.setIdleMode(self._params.neutral_mode)

        position_conversion_factor = 360 / self._params.gear_ratio
        self._encoder.setPositionConversionFactor(position_conversion_factor)
        self._encoder.setVelocityConversionFactor(position_conversion_factor / 60)

    def follow_angle(self, angle: Rotation2d):
        self._controller.setReference(angle.degrees(), rev.CANSparkMax.ControlType.kPosition)

    def reset(self):
        absolute_position = self._absolute_encoder.absolute_position - self._offset
        self._encoder.setPosition(absolute_position.degrees())

    @property
    def rotational_velocity(self) -> float:
        return self._encoder.getVelocity()

    @property
    def angle(self) -> Rotation2d:
        return Rotation2d.fromDegrees(self._encoder.getPosition())


class DummyCoaxialComponent(CoaxialDriveComponent, CoaxialAzimuthComponent):
    """Coaxial drive or azimuth component that does nothing"""

    def __init__(self, *args):
        pass

    def follow_velocity_open(self, velocity: float):
        pass

    def follow_velocity_closed(self, velocity: float):
        pass

    def reset(self):
        pass

    @property
    def velocity(self) -> float:
        return 0

    @property
    def distance(self) -> float:
        return 0

    def follow_angle(self, angle: Rotation2d):
        pass

    @property
    def rotational_velocity(self) -> float:
        return 0

    @property
    def angle(self) -> Rotation2d:
        return Rotation2d.fromDegrees(0)



import enum

import ctre.sensors
import rev
import wpilib
from wpimath.geometry import Rotation2d

from ..abstract.sensor import AbsoluteEncoder, Gyro


class AbsoluteCANCoder(AbsoluteEncoder):
    def __init__(self, id_: int | tuple[int, str]):
        super().__init__()

        # Construct the CANCoder from either a tuple of motor ID and CAN bus ID or just a motor ID
        try:
            self._encoder = ctre.sensors.CANCoder(*id_)
        except TypeError:
            self._encoder = ctre.sensors.CANCoder(id_)

        self._encoder.configAbsoluteSensorRange(ctre.sensors.AbsoluteSensorRange.Unsigned_0_to_360)

        wpilib.SmartDashboard.putData(f"Absolute CANCoder {id_}", self)

    @property
    def absolute_position(self) -> Rotation2d:
        return Rotation2d.fromDegrees(self._encoder.getAbsolutePosition())


class PigeonGyro(Gyro):
    def __init__(self, id_: int, invert: bool = False):
        super().__init__()

        self._gyro = ctre.sensors.PigeonIMU(id_)
        self.invert = invert

        wpilib.SmartDashboard.putData("Pigeon IMU", self)

    def zero_heading(self):
        self._gyro.setYaw(0)

    @property
    def heading(self) -> Rotation2d:
        yaw = self._gyro.getYaw()
        if self.invert:
            yaw = 360 - yaw
        return Rotation2d.fromDegrees(yaw)


class DummyGyro(Gyro):
    """Gyro that does nothing"""

    def __init__(self, *args):
        super().__init__()

    def zero_heading(self):
        pass

    @property
    def heading(self) -> Rotation2d:
        return Rotation2d.fromDegrees(0)


class SparkMaxEncoderType(enum.Enum):
    ANALOG = enum.auto()
    PWM = enum.auto()


class SparkMaxAbsoluteEncoder(AbsoluteEncoder):
    def __init__(self, controller: rev.CANSparkMax, encoder_type: SparkMaxEncoderType):
        """
        Absolute encoder plugged into the SPARK MAX's data port

        :param controller: SPARK MAX instance
        :param encoder_type: Type of encoder plugged in. Based on how the encoder transmits data
        """

        super().__init__()

        # Two types of absolute encoders can be plugged into the SPARK MAX data port: analog and duty cycle/PWM
        if encoder_type is SparkMaxEncoderType.ANALOG:
            self._encoder = controller.getAnalog(rev.SparkMaxAnalogSensor.Mode.kAbsolute)

            # Analog encoders output from 0V - 3.3V
            # Change from voltage to degrees
            self._encoder.setPositionConversionFactor(360 / 3.3)
        elif encoder_type is SparkMaxEncoderType.PWM:
            self._encoder = controller.getAbsoluteEncoder(rev.SparkMaxAbsoluteEncoder.Type.kDutyCycle)

            # Duty cycle encoders output from 0 to 1 by default
            # Change into degrees
            self._encoder.setPositionConversionFactor(360)

        wpilib.SmartDashboard.putData(f"Absolute Encoder {controller.getDeviceId()}", self)

    @property
    def absolute_position(self) -> Rotation2d:
        return Rotation2d.fromDegrees(self._encoder.getPosition())
